from pyramid.security import Allow


class AssertionApp:
    __name__ = "assertions"

    def __init__(self, badge):
        self.badge = badge

    def __getitem__(self, key):
        for assertion in self.badge.assertions:
            if assertion.recipient == key:
                assertion.__parent__ = self
                assertion.__name__ = assertion.recipient
                return assertion
        else:
            raise KeyError(f"Assertion {key!r} not found.")


class InvitationApp:
    __name__ = "invitations"

    def __init__(self, request):
        self.request = request

    def __getitem__(self, key):
        resource = self.request.db.get_invitation(key)
        resource.__parent__ = self
        resource.__name__ = resource.id
        return resource


class RootApp:
    __name__ = None
    __parent__ = None

    __acl__ = [
        (Allow, "group:admins", "admin"),
    ]

    def __init__(self, request):
        self.request = request

    def __getitem__(self, key):
        if key == "assertions":
            return self

        if key == "invitations":
            resource = InvitationApp(request=self.request)
            resource.__parent__ = self
            return resource

        # else

        badge = self.request.db.get_badge(key)
        if not badge:
            raise KeyError(f"No such badge {key!r}")
        resource = AssertionApp(badge=badge)
        resource.__parent__ = self
        return resource


def get_root(request):
    return RootApp(request=request)



"""This module contains generic Pyramid integration with Authlib.

It could be sent to Authlib upstream.
"""

from pyramid.httpexceptions import HTTPFound
from pyramid.threadlocal import get_current_registry


class TokenUpdate:
    """Token update event."""

    def __init__(self, name, token, refresh_token, access_token):
        """Initialize the TokenUpdate event.

        Args:
            name (str): The name of the authlib client.
            token (dict): The set of tokens the provider sent us.
            refresh_token (str): The refresh token.
            access_token (str): The access token.
        """
        self.name = name
        self.token = token
        self.refresh_token = refresh_token
        self.access_token = access_token


class PyramidIntegrationMixin:
    """Pyramid-specific framework integration."""

    def update_token(self, token, refresh_token=None, access_token=None):
        """Send an event notification on token update with Pyramid's facility."""
        event = TokenUpdate(self.name, token, refresh_token, access_token)
        registry = get_current_registry()
        registry.notify(event)

    @staticmethod
    def load_config(oauth, name, params):
        """Load the config from Pyramid's config and the json file generated by oidc-register.

        :param oauth: The OAuth app (unused).
        :param name: The provider name.
        :param params: Additional parameters (unused).

        :return: Dict configuration used to create the OAuth client.
        """
        registry = get_current_registry()
        supported_keys = (
            "client_id",
            "client_secret",
            "server_metadata_url",
            "userinfo_endpoint",
            "authorize_url",
            "access_token_url",
        )
        config = {}
        for key in supported_keys:
            try:
                config[key] = registry.settings[f"oidc.{name}.{key}"]
            except KeyError:
                continue
        return config


class PyramidAppMixin:
    """Generic methods for a Pyramid Authlib app."""

    def authorize_redirect(self, request, redirect_uri=None, **kwargs):
        """Create a HTTP Redirect for Authorization Endpoint.

        :param request: HTTP request instance from Pyramid view.
        :param redirect_uri: Callback or redirect URI for authorization.
        :param kwargs: Extra parameters to include.
        :return: A HTTP redirect response.
        """
        rv = self.create_authorization_url(redirect_uri, **kwargs)
        self.save_authorize_data(request, redirect_uri=redirect_uri, **rv)
        return HTTPFound(location=rv["url"])



"""A Fedora-specific integration of Pyramid with Authlib.

This could be part of python-fedora for other pyramid-based apps
(except there aren't any at the moment).
"""

import time

import requests
from authlib import __version__ as authlib_version
from authlib.oauth2.rfc6750 import (
    BearerTokenValidator,
    InsufficientScopeError,
    InvalidTokenError,
)
from authlib.oauth2.rfc7662 import IntrospectionToken as BaseIntrospectionToken
from packaging.version import parse as parse_version

from .oauth_1 import PyramidOAuth2App as PyramidApp


class IntrospectionToken(BaseIntrospectionToken):
    """An introspection token that implements is_expired() and is_revoked()."""

    def is_active(self):
        """Return whether the token is active."""
        return self["active"]

    def is_expired(self):
        """Return whether the token is expired."""
        if not self.is_active():
            return True
        expires_at = self.get_expires_at()
        if not expires_at:
            return None
        return int(expires_at) < time.time()

    def is_revoked(self):
        """Return whether the token is revoked."""
        return not self.is_active()


class FedoraApp(PyramidApp):
    """Fedora-specific implementation of an Authlib app.

    This class adds support for some of Ipsilon's features.
    """

    def introspect_token(self, token_string):
        """Introspect the token by calling the OpenIDC provider.

        Args:
            token_string (str): The token

        Returns:
            IntrospectionToken: The result of the token introspection.
        """
        post_data = {
            "token": token_string,
            "token_type_hint": "access_token",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }
        if self.access_token_url:
            access_token_url = self.access_token_url
        else:
            self.load_server_metadata()
            access_token_url = self.server_metadata["token_endpoint"]
        introspect_token_url = f"{access_token_url}Info"
        resp = requests.post(introspect_token_url, data=post_data)
        resp.raise_for_status()
        token_status = resp.json()
        # Add back the token itself because we'll need it later.
        token_status["access_token"] = token_string
        return IntrospectionToken(token_status)


class IntrospectTokenValidator(BearerTokenValidator):
    """Validates a token using introspection."""

    def __init__(self, app, *args, **kwargs):
        """Initialize the IntrospectTokenValidator."""
        super().__init__(*args, **kwargs)
        self.app = app

    def introspect_token(self, token_string):
        """Return the result of the app's introspection method."""
        return self.app.introspect_token(token_string)

    def authenticate_token(self, token_string):
        """Authenticate the token string and return an introspection result."""
        return self.introspect_token(token_string)

    def scope_insufficient(self, token, required_scopes):
        """Wrap the original method to support both Authlib < 1.0 and Authlib >= 1.0.

        Remove this when we only support Authlib >= 1.0.

        Args:
            token (IntrospectionToken): The token to check for insufficient scopes.
            required_scopes (list(str)): The list of required scopes.

        Returns:
            bool: Whether the token's scopes are sufficient.
        """
        if parse_version(authlib_version) >= parse_version("1.0.0"):
            return BearerTokenValidator.scope_insufficient(token.get_scope(), required_scopes)
        return super().scope_insufficient(token, required_scopes)

    def __call__(self, token_string, scopes, request, scope_operator="AND"):
        """Validate the token using introspection.

        Args:
            token_string (str): The token string.
            scopes (list): The list of required scopes.
            request (pyramid.request.Request): The Pyramid request
            scope_operator (str, optional): The operator between scopes. Defaults to "AND".

        Raises:
            InvalidTokenError: The token is invalid.
            InsufficientScopeError: The token does not have the required scopes.

        Returns:
            IntrospectedToken: The result of the token's introspection.
        """
        token = self.authenticate_token(token_string)
        self.validate_token(token, scopes, request)
        return token

    # In Authlib 1.0
    def validate_token(self, token, scopes, request):
        """Validate the token using introspection.

        Args:
            token_string (str): The token string.
            scopes (list): The list of required scopes.
            request (pyramid.request.Request): The Pyramid request
            scope_operator (str, optional): The operator between scopes. Defaults to "AND".

        Raises:
            InvalidTokenError: The token is invalid.
            InsufficientScopeError: The token does not have the required scopes.

        Returns:
            IntrospectedToken: The result of the token's introspection.
        """
        if not token:
            raise InvalidTokenError(realm=self.realm)
        if token.is_expired() or token.is_revoked():
            raise InvalidTokenError(realm=self.realm)
        if parse_version(authlib_version) < parse_version("1.0.0"):
            scopes = scopes[0]
        if self.scope_insufficient(token, scopes):
            if parse_version(authlib_version) >= parse_version("0.15.5") and parse_version(
                authlib_version
            ) < parse_version("1.0.0"):
                raise InsufficientScopeError(token_scope=token["scope"], required_scope=scopes)
            raise InsufficientScopeError()



"""Some authentication-related utilities."""

import typing

from authlib.oauth2.rfc6750 import InvalidTokenError
from pyramid.httpexceptions import HTTPFound
from pyramid.security import remember
from pyramid.settings import asbool

if typing.TYPE_CHECKING:  # pragma: no cover
    import pyramid.request.Request  # noqa: 401
    import pyramid.response.Response  # noqa: 401


def get_and_store_user(request, access_token, response):
    settings = request.registry.settings
    userinfo = request.registry.oidc.fedora.userinfo(token={"access_token": access_token})
    if "error" in userinfo:
        raise InvalidTokenError(description=userinfo["error_description"])

    nickname = userinfo["nickname"]
    if asbool(settings.get("tahrir.use_openid_email")):
        email = userinfo["email"]
        avatar = None
    else:
        email = nickname + settings.get("tahrir.email_domain")
        avatar = userinfo["email"]

    # Keep adding underscores until we get a default nickname
    # that isn't already used.
    while request.db.get_person(nickname=nickname):
        nickname += "_"

    existing = request.db.get_person(person_email=email)
    if not existing:
        request.db.add_person(email=email, nickname=nickname, avatar=avatar)
    elif existing._avatar != avatar:
        existing._avatar = avatar
        request.db.session.commit()

    # Note that they have logged in if we are installed with a newer version of
    # the db API that supports this.
    if hasattr(request.db, "note_login"):
        request.db.note_login(person_email=email)

    headers = remember(request, email)
    # response = HTTPFound(location=request.session.get('came_from', '/'))
    response.headerlist.extend(headers)
    return response


def get_final_redirect(request: "pyramid.request.Request"):
    """Get the URL that the user should be redirected to after logging in.

    Args:
        request (pyramid.request.Request): the current request.

    Returns:
        HTTPFound: An HTTP 302 response redirecting to the right URL.
    """
    came_from = request.session.get("came_from", request.route_path("home"))
    request.session.pop("came_from", None)

    # Mitigate "Covert Redirect"
    if not came_from.startswith(request.host_url):
        came_from = request.route_path("home")
    # Don't redirect endlessly to the login view
    if came_from.startswith(request.route_url("login")):
        came_from = request.route_path("home")

    return HTTPFound(location=came_from)



# Copyright (c) 2019 Red Hat, Inc.

"""Test tahrir.utils."""

import unittest

from tahrir import utils


class TestSingularize(unittest.TestCase):
    """Test the singularize() function."""

    def test_singularize_value_1(self):
        """Test that the trailing letter is returned when value is 1."""
        self.assertEqual(utils.singularize("cats", 1), "cat")


class TestStrToBytes(unittest.TestCase):
    """Test the str_to_bytes() function."""

    def test_str_to_bytes_str(self):
        self.assertEqual(utils.str_to_bytes("foo"), b"foo")

    def test_str_to_bytes_bytes(self):
        self.assertEqual(utils.str_to_bytes(b"foo"), b"foo")

    def test_str_to_bytes_other(self):
        self.assertEqual(utils.str_to_bytes(["list"]), ["list"])



from pkg_resources import get_distribution
from pyramid.events import (
    BeforeRender,
    subscriber,
)
from pyramid.settings import asbool

import tahrir.views


@subscriber(BeforeRender)
def inject_globals(event):
    """Injects global variables into every template after the view
    is processed but before the template is rendered."""

    # request is available in every template, and we can just pull it
    # in like so...
    request = event["request"]

    settings = request.registry.settings

    # ... and then set a couple global variables that will be available
    # in every template, so we don't have to pass them through the
    # dict returned by the view every time!
    event["title"] = settings["tahrir.title"]
    event["base_url"] = settings["tahrir.base_url"]
    event["theme_name"] = settings.get("tahrir.theme_name", "tahrir")

    event["tahrir_version"] = get_distribution("tahrir").version
    event["tahrir_api_version"] = get_distribution("tahrir-api").version

    event["logged_in"] = request.authenticated_userid
    person = request.db.get_person(event["logged_in"])
    event["logged_in_person"] = person

    event["footer"] = tahrir.views.load_docs(request, "footer")

    event["twitter"] = asbool(settings.get("tahrir.social.twitter"))
    event["twitter_user_text"] = settings.get("tahrir.social.twitter_user_text")
    event["twitter_user_hash"] = settings.get("tahrir.social.twitter_user_hash")

    event["auth_principals"] = request.effective_principals



import datetime
import math
import urllib.parse
from hashlib import sha256

import dateutil.relativedelta
import dogpile.cache
import dogpile.cache.util
import pyramid.threadlocal

libravatar = None
try:
    import libravatar
except ImportError:
    pass


cache = dogpile.cache.make_region(
    key_mangler=lambda x: dogpile.cache.util.sha1_mangle_key(str_to_bytes(x))
)


def generate_badge_yaml(postdict):
    return (
        "%YAML 1.2\n"
        "---\n"
        "\n"
        "# This is some metadata about the badge.\n"
        "name:           " + postdict.get("badge-name", default="") + "\n"
        "description:    " + postdict.get("badge-description", default="") + "\n"
        "creator:        " + postdict.get("badge-creator", default="") + "\n"
        "\n"
        "# This is a link to the discussion about adopting this as\n"
        "a for-real badge\n"
        "discussion:     " + postdict.get("discussion", default="") + "\n"
        "\n"
        "# A link to the image for the badge.\n"
        "image_url:      " + postdict.get("image", default="") + "\n"
        "\n"
        "# The issuer.\n"
        "issuer_id:      " + postdict.get("issuer", default="") + "\n"
        "\n"
        "# We'll perform our more costly check (defined below)\n"
        "# only when we receive messages that match this trigger.\n"
        "trigger:\n"
        "  topic:        " + postdict.get("trigger-topic", default="") + "\n"
        "\n"
        "# Once the check has been triggered, this defines what we\n"
        "# actually check.\n"
        "criteria:       " + postdict.get("criteria", default="") + "\n"
        "(This section is under construction.)"
    )


@cache.cache_on_arguments()
def get_avatar(email: str, size):
    request = pyramid.threadlocal.get_current_request()
    default = request.registry.settings.get("tahrir.default_avatar", "retro")

    query = {
        "s": size,
        "d": default,
    }

    if size == "responsive":
        # Make it big so we can downscale it as we please
        query["s"] = 312

    query = urllib.parse.urlencode(query)

    # Use md5 for emails, and sha256 for openids.
    # We're really using openids, so...
    # hash = md5(email).hexdigest()
    hash = sha256(email.encode("utf-8")).hexdigest()

    # TODO This next line is temporary and can be removed.  We do
    # libravatar ourselves here by hand to avoid pyDNS issues on epel6.
    # Once those are resolved we can use pylibravatar again.
    return f"https://seccdn.libravatar.org/avatar/{hash}?{query}"

    gravatar_url = f"https://secure.gravatar.com/avatar/{hash}?{query}"

    if libravatar:
        return libravatar.libravatar_url(
            email=email,
            size=size,
            default=gravatar_url,
        )
    else:
        return gravatar_url


def singularize(term, value):
    """Strip the 's' off of plural words to dumbly singularize them."""
    if value == 1:
        return term[:-1]
    else:
        return term


def relative_time(value: datetime.datetime):
    SHORT_DENOMINATIONS = {
        "years": "yrs",
        "months": "mons",
        "days": "days",
        "hours": "hrs",
        "minutes": "mins",
        "seconds": "secs",
    }
    then_in_seconds = value.timestamp()
    now_in_seconds = datetime.datetime.now(datetime.timezone.utc).timestamp()
    delta = now_in_seconds - then_in_seconds

    if delta > 0:
        suffix = "ago"
    else:
        suffix = "from now"

    # time_strings = []
    rd = dateutil.relativedelta.relativedelta(seconds=math.fabs(delta))
    denominations = ["years", "months", "days", "hours", "minutes", "seconds"]
    for denomination in denominations:
        value = getattr(rd, denomination, 0)
        if value:
            return "%d %s %s" % (
                value,
                singularize(SHORT_DENOMINATIONS[denomination], value),
                suffix,
            )

    return "just now"


def str_to_bytes(input):
    """If input is a unicode string, encodes it and returns the result.

    Otherwise just passes it through. Needed to deal with dogpile key mangling.
    """
    if isinstance(input, str):
        input = input.encode("utf-8")
    return input



import codecs
import os
import random
import types
from datetime import date, datetime, timedelta, timezone
from decimal import ROUND_UP, Decimal
from io import BytesIO

import docutils.examples
import markupsafe
import qrcode as qrcode_module
import sqlalchemy as sa
import tahrir_api.model as m
from feedgen.feed import FeedGenerator
from pyramid.httpexceptions import (
    HTTPForbidden,
    HTTPFound,
    HTTPGone,
    HTTPMethodNotAllowed,
    HTTPNotFound,
)
from pyramid.response import Response
from pyramid.view import view_config
from tahrir_api.utils import convert_name_to_id

from tahrir.utils import generate_badge_yaml, get_avatar


def _get_user(request, id_or_nickname):
    """Attempt to get a user by their id or nickname, returning None if
    we fail."""
    user = request.db.get_person(nickname=id_or_nickname)

    if user:
        return user
    else:
        try:
            # We cast user_id to an integer so that Postgres doesn't
            # get upset about comparing what is potentially a string
            # to an integer column.
            return request.db.get_person(id=int(id_or_nickname))
        except ValueError:
            return None


def _get_team(request, team_id):
    """Get team for the given team_id"""
    team = request.db.get_team(team_id=team_id)

    return team


def _get_user_badge_info(request, user):
    """Returns a dictionary of the user badge information"""

    # Get user badges.
    user_badges = [a.badge for a in user.assertions]

    # Sort user badges by id.
    user_badges = sorted(user_badges, key=lambda badge: badge.id)

    # Get total number of unique badges in the system.
    count_total_badges = request.db.get_all_badges().count()

    # Get percentage of badges earned.
    try:
        percent_earned = (float(len(user_badges)) / float(count_total_badges)) * 100
    except ZeroDivisionError:
        percent_earned = 0

    # Get rank. (same code found in leaderboard view function)
    rank = user.rank or 0
    user_count = request.db.session.query(m.Person).filter(m.Person.opt_out.is_(False)).count()

    try:
        percentile = (
            Decimal(float(rank) / float(user_count)).quantize(Decimal(".0001"), rounding=ROUND_UP)
            * 100
        )
    except ZeroDivisionError:
        percentile = 0

    return dict(
        user_badges=user_badges,
        count_total_badges=count_total_badges,
        percent_earned=percent_earned,
        rank=rank,
        user_count=user_count,
        percentile=percentile,
    )


@view_config(route_name="award", renderer="string")
def award(request):
    if not request.POST:
        return HTTPMethodNotAllowed()

    token = request.session.get_csrf_token()
    if token != request.POST["csrf_token"]:
        raise HTTPForbidden("CSRF token did not match")

    agent = request.db.get_person(request.authenticated_userid)
    if not agent:
        raise HTTPForbidden()

    badge_id = request.POST.get("badge_id")
    badge = request.db.get_badge(badge_id)
    if not badge:
        raise HTTPNotFound(f"No such badge {badge_id!r}")

    if not badge.authorized(agent):
        raise HTTPForbidden(f"Unauthorized for {badge_id!r}")

    nickname = request.POST.get("nickname")
    user = request.db.get_person(nickname=nickname)
    if not user:
        raise HTTPNotFound(f"No such user {nickname!r}")

    if user.opt_out:
        raise HTTPNotFound(f"No such user {nickname!r}")

    # OK
    request.db.add_assertion(badge.id, user.email, None)

    return HTTPFound(location=request.route_url("badge_rss", id=badge.id))


@view_config(route_name="invite", renderer="string")
def invite(request):
    if not request.POST:
        return HTTPMethodNotAllowed()

    token = request.session.get_csrf_token()
    if token != request.POST["csrf_token"]:
        raise HTTPForbidden("CSRF token did not match")

    agent = request.db.get_person(request.authenticated_userid)
    if not agent:
        raise HTTPForbidden()

    badge_id = request.POST.get("badge_id")
    badge = request.db.get_badge(badge_id)
    if not badge:
        raise HTTPNotFound(f"No such badge {badge_id!r}")

    if not badge.authorized(agent):
        raise HTTPForbidden(f"Unauthorized for {badge_id!r}")

    try:
        fmt = "%Y-%m-%d %H:%M"
        expires_on = datetime.strptime(request.POST.get("expires-on"), fmt)
    except ValueError:
        expires_on = None  # Will default to 1 hour from now

    # OK
    request.db.add_invitation(badge.id, expires_on=expires_on, created_by_email=agent.email)

    return HTTPFound(location=request.route_url("user", id=agent.id))


@view_config(route_name="add_tag", renderer="string", permission="admin")
def add_tag(request):
    if not request.POST:
        return HTTPMethodNotAllowed()

    token = request.session.get_csrf_token()
    if token != request.POST["csrf_token"]:
        raise HTTPForbidden("CSRF token did not match")

    agent = request.db.get_person(request.authenticated_userid)
    if not agent:
        raise HTTPForbidden()

    badge_id = request.POST.get("badge_id")
    badge = request.db.get_badge(badge_id)
    if not badge:
        raise HTTPNotFound(f"No such badge {badge_id!r}")

    tags = request.POST.get("tags", "")
    new_tags = [tag.strip() for tag in tags.strip().split(",") if tag.strip()]
    originals = [tag.strip() for tag in badge.tags.split(",") if tag.strip()]
    badge.tags = ",".join(set(originals + new_tags)) + ","
    request.db.session.flush()

    return HTTPFound(location=request.route_url("badge", id=badge.id))


@view_config(route_name="admin", renderer="admin.mak", permission="admin")
def admin(request):
    # settings = request.registry.settings

    # TODO: Check if I even need this anymore... leaving for now.
    request.session["came_from"] = request.route_url("admin")

    if request.authenticated_userid:
        awarded_assertions = request.db.get_assertions_by_email(request.authenticated_userid)
    else:
        awarded_assertions = None

    # Handle any admin actions. These are done through POSTS via the
    # HTML forms on the admin panel.
    if request.POST:
        token = request.session.get_csrf_token()
        if token != request.POST["csrf_token"]:
            raise HTTPForbidden("CSRF token did not match")

        if request.POST.get("add-team"):
            team_name = request.POST.get("team-name")
            team_id = convert_name_to_id(team_name)
            if not request.db.team_exists(team_id=team_id):
                request.db.create_team(name=team_name)
                request.session.flash(f"You created a team with name {team_name}")
            else:
                request.session.flash(f"Team with name {team_name} already exists.")

        elif request.POST.get("add-series"):
            data = request.POST
            series_name = data.get("series-name")
            series_id = convert_name_to_id(series_name)
            if not request.db.series_exists(series_id):
                description = data.get("series-description")
                tags = data.get("series-tags")
                team_id = data.get("series-team-id")

                request.db.create_series(
                    series_id=series_id,
                    name=series_name,
                    desc=description,
                    tags=tags,
                    team_id=team_id,
                )
                request.session.flash(f"You created a series with name {series_name}")
            else:
                request.session.flash(f"Series with name {series_name} already exists.")

        elif request.POST.get("add-milestone"):
            data = request.POST
            series_id = data.get("milestone-series-id")
            badge_id = data.get("milestone-badge-id")
            if not request.db.milestone_exists_for_badge_series(badge_id, series_id):
                position = data.get("milestone-position")
                request.db.create_milestone(
                    position=position, series_id=series_id, badge_id=badge_id
                )
                request.session.flash(
                    f"You add badge {badge_id} as milestone in series {series_id}"
                )
            else:
                request.session.flash(
                    f"Badge {badge_id} already added as milestone in series {series_id}"
                )

        elif request.POST.get("add-person"):
            # Email is a required field on the HTML form.
            # Add a Badge to the DB.
            email = request.POST.get("person-email")
            if not request.db.person_exists(email=email):
                request.db.add_person(
                    email,
                    nickname=request.POST.get("person-nickname"),
                )
                request.session.flash(
                    "You added a person with email {}".format(request.POST.get("person-email"))
                )
            else:
                request.session.flash(f"Person with email {email} already exists.")
        elif request.POST.get("add-badge"):
            name = request.POST.get("badge-name")
            idx = convert_name_to_id(name)
            if not request.db.badge_exists(idx):
                # Add a Badge to the DB.
                request.db.add_badge(
                    name,
                    request.POST.get("badge-image"),
                    request.POST.get("badge-description"),
                    request.POST.get("badge-criteria"),
                    request.POST.get("badge-issuer"),
                    request.POST.get("badge-tags"),
                )
                request.session.flash(f"You added a badge with name {name}")
            else:
                request.session.flash(f"Badge with id {idx} already exists.")

        elif request.POST.get("add-invitation"):
            # Add an Invitation to the DB.
            try:
                created_on = datetime.strptime(
                    request.POST.get("invitation-created"), "%Y-%m-%d %H:%M"
                )
            except ValueError:
                created_on = None  # Will default to datetime.now(timezone.utc)

            try:
                expires_on = datetime.strptime(
                    request.POST.get("invitation-expires"), "%Y-%m-%d %H:%M"
                )
            except ValueError:
                expires_on = None  # Will default to datettime.now(timezone.utc)

            try:
                request.db.add_invitation(
                    request.POST.get("invitation-badge-id"),
                    created_on=created_on,
                    expires_on=expires_on,
                    created_by_email=request.POST.get("invitation-issuer-email"),
                )
            except ValueError as e:
                request.session.flash(str(e))
            else:
                request.session.flash(
                    "You added an invitation for badge {}".format(
                        request.POST.get("invitation-badge-id")
                    )
                )
        elif request.POST.get("add-issuer"):
            origin = request.POST.get("issuer-origin")
            name = request.POST.get("issuer-name")
            if not request.db.issuer_exists(origin, name):
                # Add an Issuer to the DB.
                request.db.add_issuer(
                    origin, name, request.POST.get("issuer-org"), request.POST.get("issuer-contact")
                )
                request.session.flash(
                    "You added an issuer with the name {}".format(request.POST.get("issuer-name"))
                )
            else:
                request.session.flash(
                    f"Issuer with origin {origin} and name {name} already exists."
                )

        elif request.POST.get("add-assertion"):
            idx = request.POST.get("assertion-badge-id")
            email = request.POST.get("assertion-person-email")
            if not request.db.assertion_exists(idx, email):
                # Add an Assertion to the DB.
                try:
                    issued_on = datetime.strptime(
                        request.POST.get("assertion-issued-on"), "%Y-%m-%d %H:%M"
                    )
                except ValueError:
                    issued_on = None  # Will default to datetime.now(timezone.utc)

                request.db.add_assertion(
                    request.POST.get("assertion-badge-id"),
                    request.POST.get("assertion-person-email"),
                    issued_on,
                )
                request.session.flash(
                    f"You awarded {request.POST.get('assertion-badge-id')} "
                    f"to {request.POST.get('assertion-person-email')}"
                )
            else:
                request.session.flash(f"User with email {email} already has badge {idx}.")

        elif request.POST.get("add-authorization"):
            idx = request.POST.get("authorization-badge-id")
            email = request.POST.get("authorization-person-email")

            if not request.db.authorization_exists(idx, email):
                request.db.add_authorization(idx, email)
                request.session.flash(f"You authorized {email} to issue {idx}")
            else:
                request.session.flash(f"{email} is already authorized to issue {idx}.")

    return dict(
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
        issuers=request.db.get_all_issuers().all(),
    )


@view_config(route_name="heartbeat", renderer="string")
def heartbeat(request):
    # A NOOP for haproxy httpchk to ping
    n = request.db.get_all_badges().count()
    return "OK: (%i badges in the system)" % n


@view_config(route_name="home", renderer="index.mak")
def index(request):

    n = 5  # n is the number of items displayed in each column.

    # set came_from so we can get back home after openid auth.
    request.session["came_from"] = request.route_url("home")

    latest_awards = (
        request.db.get_all_assertions()
        .join(m.Person)
        .filter(m.Person.opt_out.is_(False))
        .order_by(sa.desc(m.Assertion.issued_on))
        .limit(n)
        .all()
    )

    start = get_start_week()
    stop = start + timedelta(days=6)
    weekly_leaders = request.db._make_leaderboard(
        start=start,
        stop=stop,
    )

    now = datetime.now(timezone.utc)
    start = date(now.year, now.month, 1)
    if now.month == 12:
        stop = date(now.year + 1, 1, 1) - timedelta(days=1)
    else:
        stop = date(now.year, now.month + 1, 1) - timedelta(days=1)
    monthly_leaders = request.db._make_leaderboard(
        start=start,
        stop=stop,
    )

    return dict(
        auth_principals=request.effective_principals,
        latest_awards=latest_awards,
        weekly_leaders=weekly_leaders,
        monthly_leaders=monthly_leaders,
        n=n,
    )


@view_config(context=types.FunctionType, permission="admin")
def action(context, request):

    # Do the action
    context()

    return HTTPFound(location=request.route_url("home"))


@view_config(context=m.Invitation, name="claim")
def invitation_claim(request):
    """Action that awards a person a badge after scanning a qrcode."""

    # settings = request.registry.settings

    if request.context.expires_on < datetime.now():
        return HTTPGone("That invitation is expired.")

    if not request.authenticated_userid:
        request.session["came_from"] = request.resource_url(request.context, "claim")
        return HTTPFound(location=request.route_url("login"))

    person = request.db.get_person(person_email=request.authenticated_userid)

    # Check to see if the user already has the badge.
    if request.context.badge in [a.badge for a in person.assertions]:
        request.session.flash("You already have " + request.context.badge_id + " badge")
        return HTTPFound(location=request.route_url("home"))

    request.db.add_assertion(request.context.badge_id, person.email, datetime.now(timezone.utc))

    # TODO -- return them to a page that auto-exports their badges.
    request.session.flash("You have earned " + request.context.badge_id + " badge")
    return HTTPFound(location=request.route_url("home"))


@view_config(context=m.Invitation, name="qrcode")
def invitation_qrcode(request):
    """Returns a raw dummy qrcode through to the user."""

    if request.context.expires_on < datetime.now():
        return HTTPGone("That invitation is expired.")

    target = request.resource_url(request.context, "claim")
    img = qrcode_module.make(target)
    bytestream = BytesIO()
    img.save(bytestream)
    return Response(
        body=bytestream.getvalue(),
        content_type="image/png",
    )


@view_config(route_name="leaderboard", renderer="leaderboard.mak")
def leaderboard(request):
    """Render a top users view."""

    user, awarded_assertions = None, None

    if request.authenticated_userid:
        user = request.db.get_person(person_email=request.authenticated_userid)

    query = (
        request.db.session.query(m.Person)
        .order_by(
            m.Person.rank,
            m.Person.created_on,
        )
        .filter(m.Person.opt_out.is_(False))
    )

    leaderboard = query.filter(m.Person.rank.isnot(None)).all()
    # Get total user count.
    user_count = len(leaderboard)
    leaderboard.extend(query.filter(m.Person.rank.is_(None)).all())

    user_to_rank = request.db._make_leaderboard()

    if user:
        awarded_assertions = user.assertions
        rank = user.rank or 0
        idx = rank - 1

        # Handle the case of leaderboard[-2:2] which will be [] always.
        if idx < 2:
            idx = 2

        competitors = leaderboard[(idx - 2) : (idx + 3)]

        try:
            percentile = (float(rank) / float(user_count)) * 100
        except ZeroDivisionError:
            percentile = 0
    else:
        awarded_assertions = None
        rank = None
        competitors = None
        percentile = None

    return dict(
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
        top_persons_sorted=leaderboard,
        rank=rank,
        user_count=user_count,
        percentile=percentile,
        competitors=competitors,
        user_to_rank=user_to_rank,
    )


@view_config(route_name="leaderboard_json", renderer="json")
@view_config(route_name="rank_json", renderer="json")
def leaderboard_json(request):
    """Render a top-users JSON dump."""

    limit = int(request.params.get("limit", 25))
    user_id = request.matchdict.get("id")
    user = None
    if user_id:
        user = _get_user(request, user_id)

    query = (
        request.db.session.query(m.Person)
        .order_by(
            m.Person.rank,
            m.Person.created_on,
        )
        .filter(m.Person.opt_out.is_(False))
    )

    leaderboard = query.filter(m.Person.rank.isnot(None)).all()
    # Get total user count.
    # user_count = len(leaderboard)
    leaderboard.extend(query.filter(m.Person.rank.is_(None)).all())

    user_to_rank = request.db._make_leaderboard()

    if user:
        rank = user.rank or 0
        idx = rank - 1

        # Handle the case of leaderboard[-2:2] which will be [] always.
        if idx < 2:
            idx = 2

        leaderboard = leaderboard[(idx - 2) : (idx + 3)]
    else:
        leaderboard = leaderboard[:limit]

    ret = [
        {**user_to_rank[p], **{"nickname": p.nickname}} for p in leaderboard if p in user_to_rank
    ]

    return {"leaderboard": ret}


@view_config(route_name="about", renderer="about.mak")
def about(request):
    return dict(content=load_docs(request, "about"), auth_principals=request.effective_principals)


@view_config(route_name="explore", renderer="explore.mak")
def explore(request):

    # Check if a search has been done, and if so, show
    # search results.
    search_query = None
    search_results = dict()  # name: link
    if request.POST:
        token = request.session.get_csrf_token()
        if token != request.POST["csrf_token"]:
            raise HTTPForbidden("CSRF token did not match")

        if request.POST.get("badge-search"):
            # badge-query is a required field on the template form.
            search_query = request.POST.get("badge-query")
            matching_results = (
                request.db.get_all_badges()
                .filter(
                    sa.func.lower(m.Badge.name).like("%" + search_query + "%")
                    | sa.func.lower(m.Badge.description).like("%" + search_query + "%")
                    | sa.func.lower(m.Badge.tags).like("%" + search_query + "%")
                )
                .all()
            )
            for r in matching_results:
                search_results[r.name] = request.route_url("badge", id=r.id)
        elif request.POST.get("person-search"):
            # person-query is a required field on the template form.
            search_query = request.POST.get("person-query")
            matching_results = (
                request.db.get_all_persons()
                .filter(
                    (m.Person.nickname.like("%" + search_query + "%"))
                    & (m.Person.opt_out.is_(False))
                )
                .all()
            )
            for r in matching_results:
                search_results[r.nickname] = request.route_url("user", id=r.nickname)
        elif request.POST.get("tag-search"):
            # tag-query is a required field on the template form.
            search_query = request.POST.get("tag-query")
            if request.POST.get("tag-match-all"):
                return HTTPFound(location=request.route_url("tags", tags=search_query, match="all"))
            else:
                return HTTPFound(location=request.route_url("tags", tags=search_query, match="any"))

    # Get awarded assertions.
    if request.authenticated_userid:
        awarded_assertions = request.db.get_assertions_by_email(request.authenticated_userid)
    else:
        awarded_assertions = None

    # Get some random badges (for discovery).
    try:
        random_badges = random.sample(request.db.get_all_badges().all(), 5)
    except ValueError:  # the sample is probably larger than the population
        random_badges = request.db.get_all_badges().all()

    # Get some random persons (for discovery).
    try:
        random_persons = random.sample(
            request.db.get_all_persons().filter(m.Person.opt_out.is_(False)).all(), 5
        )
    except ValueError:  # the sample is probably larger than the population
        random_persons = request.db.get_all_persons().filter(m.Person.opt_out.is_(False)).all()

    return dict(
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
        random_badges=random_badges,
        random_persons=random_persons,
        search_results=search_results,
        search_query=search_query,
    )


@view_config(route_name="explore_badges", renderer="explore_badges.mak")
def explore_badges(request):

    all_badges = request.db.get_all_badges().all()

    newest_badges = sorted(
        request.db.get_all_badges().all(), key=lambda badge: badge.created_on, reverse=True
    )[:20]

    # Get awarded assertions.
    if request.authenticated_userid:
        awarded_assertions = request.db.get_assertions_by_email(request.authenticated_userid)
    else:
        awarded_assertions = None

    return dict(
        all_badges=all_badges,
        newest_badges=newest_badges,
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
    )


@view_config(route_name="explore_badges_rss")
def explore_badges_rss(request):
    """Render rss feed for the latest badges."""

    newest_badges = sorted(
        request.db.get_all_badges().all(), key=lambda badge: badge.created_on, reverse=True
    )[:20]

    feed = FeedGenerator()
    feed.title("Newest badges Feed")
    feed.link(href=request.route_url("explore_badges_rss"), rel="self")
    feed.subtitle("Latest badges of the application")
    feed.language = "en"

    description_template = "<img src='%s' alt='%s' />%s"

    for badge in newest_badges:
        url = request.route_url("badge", id=badge.id)
        entry = feed.add_entry()
        entry.title(f"New badge: {badge.name} !")
        entry.link(href=url)
        pubdate = badge.created_on.replace(tzinfo=timezone.utc)
        entry.published(pubdate)
        entry.description(
            description_template
            % (
                badge.image,
                badge.name,
                badge.description,
            )
        )

    return Response(
        body=feed.rss_str(pretty=True),
        content_type="application/rss+xml",
        charset="utf-8",
    )


@view_config(route_name="badge_full", renderer="badge_full.mak")
@view_config(route_name="badge", renderer="badge.mak")
def badge(request):
    """Render badge info page."""
    # Get the badge to render info about.
    badge_id = request.matchdict.get("id")
    badge = request.db.get_badge(badge_id)

    # if the badge isn't found, raise a 404
    if not badge:
        raise HTTPNotFound(f"No such badge {badge_id!r}")

    # Get awarded assertions.
    if request.authenticated_userid:
        awarded_assertions = request.db.get_assertions_by_email(request.authenticated_userid)
    else:
        awarded_assertions = []

    # Get badge statistics.
    # TODO: Perhaps abstract these statistics methods away somewhere?
    try:
        times_awarded = len(request.db.get_assertions_by_badge(badge_id))
        last_awarded = (
            request.db.get_all_assertions()
            .filter(sa.func.lower(m.Assertion.badge_id) == sa.func.lower(badge_id))
            .order_by(sa.desc(m.Assertion.issued_on))
            .limit(1)
            .one()
        )
        last_awarded_person = request.db.get_person(id=last_awarded.person_id)

        first_awarded = (
            request.db.get_all_assertions()
            .filter(sa.func.lower(m.Assertion.badge_id) == sa.func.lower(badge_id))
            .order_by(sa.asc(m.Assertion.issued_on))
            .limit(1)
            .one()
        )
        first_awarded_person = request.db.get_person(id=first_awarded.person_id)

        percent_earned = float(times_awarded) / float(len(request.db.get_all_persons().all()))

        # This is a list of assertions for this badge.
        badge_assertions = (
            request.db.get_all_assertions().filter(m.Assertion.badge_id == badge.id).all()
        )
    except sa.orm.exc.NoResultFound:  # This badge has never been awarded.
        times_awarded = 0
        last_awarded = None
        last_awarded_person = None
        first_awarded = None
        first_awarded_person = None
        badge_assertions = None
        percent_earned = 0
    # Percent of people who have earned this badge

    # Get badge description HTML from RST.
    # Note: this html_body function wraps the output
    # in a <divclass="document">.
    badge_description_html = docutils.examples.html_body(badge.description)

    return dict(
        badge=badge,
        badge_description_html=badge_description_html,
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
        times_awarded=times_awarded,
        last_awarded=last_awarded,
        last_awarded_person=last_awarded_person,
        first_awarded=first_awarded,
        first_awarded_person=first_awarded_person,
        badge_assertions=badge_assertions,
        percent_earned=percent_earned,
    )


def _badge_json_generator(request, badge, withasserts=True):
    if not withasserts:
        return {
            "id": badge.id,
            "name": badge.name,
            "description": badge.description,
            "image": badge.image,
            "tags": badge.tags,
        }

    try:
        # Fixme -- not sure if this works -- need to check it out.
        assertions = sorted(badge.assertions, key=lambda b: b.issued_on)

        times_awarded = len(badge.assertions)

        percent_earned = float(times_awarded) / float(request.db.get_all_persons().count())

        if assertions:
            last_awarded = assertions[-1]
            last_awarded_person = last_awarded.person

            first_awarded = assertions[0]
            first_awarded_person = first_awarded.person
        else:
            last_awarded = None
            last_awarded_person = None
            first_awarded = None
            first_awarded_person = None

    except sa.orm.exc.NoResultFound:  # This badge has never been awarded.
        times_awarded = 0
        last_awarded = None
        last_awarded_person = None
        first_awarded = None
        first_awarded_person = None
        percent_earned = 0

    if last_awarded:
        last_awarded = float(last_awarded.issued_on.strftime("%s"))

    if last_awarded_person:
        last_awarded_person = last_awarded_person.nickname

    if first_awarded:
        first_awarded = float(first_awarded.issued_on.strftime("%s"))

    if first_awarded_person:
        first_awarded_person = first_awarded_person.nickname

    if percent_earned:
        percent_earned *= 100

    return {
        "id": badge.id,
        "name": badge.name,
        "description": badge.description,
        "times_awarded": times_awarded,
        "last_awarded": last_awarded,
        "last_awarded_person": last_awarded_person,
        "first_awarded": first_awarded,
        "first_awarded_person": first_awarded_person,
        "percent_earned": percent_earned,
        "image": badge.image,
        "tags": badge.tags,
    }


@view_config(route_name="badge_json", renderer="json")
def badge_json(request):
    """Render badge JSON dump."""
    # Get the badge to render info about.
    badge_id = request.matchdict.get("id")
    badge = request.db.get_badge(badge_id)

    # if the badge isn't found, raise a 404
    if not badge:
        request.response.status = "404 Not Found"
        return {"error": "No such badge exists."}

    return _badge_json_generator(request, badge)


@view_config(route_name="badge_rss")
def badge_rss(request):
    """Render per-badge rss."""

    badge_id = request.matchdict.get("id")
    badge = request.db.get_badge(badge_id)

    if not badge:
        raise HTTPNotFound(f"No such badge {badge_id!r}")

    # this gives us the assertions sorted *earliest first*. feedgen's
    # default when adding entries is to prepend - put the new item at
    # the top of the feed. so as we iterate over this and add items to
    # the feed, we add each newer assertion to the front of the feed
    sorted_assertions = sorted(badge.assertions, key=lambda x: x.issued_on)

    feed = FeedGenerator()
    feed.title(f"Badges Feed for {badge.name}")
    feed.link(href=request.route_url("badge", id=badge.id), rel="self")
    feed.subtitle(f"Latest recipients of the badge {badge.name}")
    feed.language("en")

    description_template = "<img src='%s' alt='%s' />%s"

    for assertion in sorted_assertions:
        url = request.route_url("user", id=assertion.person.nickname or assertion.person.id)
        entry = feed.add_entry()
        entry.title(assertion.person.nickname)
        entry.link(href=url)
        pubdate = assertion.issued_on.replace(tzinfo=timezone.utc)
        entry.published(pubdate)
        entry.description(
            description_template
            % (
                get_avatar(assertion.person.avatar, 128),
                assertion.person.nickname,
                assertion.person.nickname,
            )
        )

    return Response(
        body=feed.rss_str(pretty=True),
        content_type="application/rss+xml",
        charset="utf-8",
    )


@view_config(route_name="badge_stl", renderer="thingiview.mak")
def badge_stl(request):
    """Render per-badge stl."""

    badge_id = request.matchdict.get("id")
    badge = request.db.get_badge(badge_id)

    if not badge:
        raise HTTPNotFound(f"No such badge {badge_id!r}")

    if not badge.stl:
        raise HTTPNotFound("Badge has no stl file.")

    return dict(
        badge=badge,
        auth_principals=request.effective_principals,
    )


@view_config(route_name="user_rss")
def user_rss(request):
    """Render per-user rss."""

    user_id = request.matchdict.get("id")
    user = _get_user(request, user_id)

    if not user:
        raise HTTPNotFound(f"No such user {user_id!r}")

    if user.opt_out and user.email != request.authenticated_userid:
        raise HTTPNotFound(f"User {user_id!r} has opted out.")

    # this gives us the assertions sorted *earliest first*. feedgen's
    # default when adding entries is to prepend - put the new item at
    # the top of the feed. so as we iterate over this and add items to
    # the feed, we add each newer assertion to the front of the feed
    sorted_assertions = sorted(user.assertions, key=lambda x: x.issued_on)

    feed = FeedGenerator()
    feed.title(f"Badges Feed for {user.nickname}")
    feed.link(href=request.route_url("user", id=user.nickname or user.id), rel="self")
    feed.subtitle(f"The latest Fedora Badges obtained by {user.nickname}")
    feed.language("en")

    description_template = "<img src='%s' alt='%s'/>%s -- %s"

    for assertion in sorted_assertions:
        entry = feed.add_entry()
        entry.title(assertion.badge.name)
        entry.link(href=request.route_url("badge", id=assertion.badge.id))
        pubdate = assertion.issued_on.replace(tzinfo=timezone.utc)
        entry.published(pubdate)
        entry.description(
            description_template
            % (
                assertion.badge.image,
                assertion.badge.name,
                assertion.badge.name,
                assertion.badge.description,
            )
        )

    return Response(
        body=feed.rss_str(pretty=True),
        content_type="application/rss+xml",
        charset="utf-8",
    )


@view_config(route_name="user", renderer="user.mak")
def user(request):
    """Render user info page."""

    # Grab a boolean out of the config
    # settings = request.registry.settings

    # Get awarded assertions.
    if request.authenticated_userid:
        awarded_assertions = request.db.get_assertions_by_email(request.authenticated_userid)
    else:
        awarded_assertions = None

    user_id = request.matchdict.get("id")
    user = _get_user(request, user_id)

    history_limit = int(request.params.get("history_limit", 10))

    if not user:
        raise HTTPNotFound(f"No such user {user_id!r}")

    if user.opt_out and user.email != request.authenticated_userid:
        raise HTTPNotFound(f"User {user_id!r} has opted out.")

    if request.POST:

        token = request.session.get_csrf_token()
        if token != request.POST["csrf_token"]:
            raise HTTPForbidden("CSRF token did not match")

        # Authz check
        if request.authenticated_userid != user.email:
            raise HTTPForbidden("Unauthorized")

        person = request.db.get_person(person_email=request.authenticated_userid)
        if person is None:
            raise HTTPNotFound(f"Person with email {request.authenticated_userid} not found")

        if request.POST.get("deactivate-account"):
            person.opt_out = True
        elif request.POST.get("reactivate-account"):
            person.opt_out = False

    # Get invitations the user has created.
    invitations = [i for i in request.db.get_invitations(user.id) if i.expires_on > datetime.now()]

    user_info = dict(
        user=user,
        invitations=invitations,
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
        history_limit=history_limit,
    )

    user_info.update(_get_user_badge_info(request, user))

    return user_info


def _user_json_generator(request, user):
    """Generates a json of user data"""
    user_info = _get_user_badge_info(request, user)

    assertions = []
    for assertion in user.assertions:
        issued = {"issued": float(assertion.issued_on.strftime("%s"))}
        _badged = _badge_json_generator(request, assertion.badge, withasserts=False)
        assertions.append({**issued, **_badged})

    return {
        "user": user.nickname,
        "avatar": get_avatar(user.avatar, int(request.GET.get("size", 100))),
        "percent_earned": user_info["percent_earned"],
        "assertions": assertions,
        "percentile": str(user_info["percentile"]),
        "rank": user_info["rank"],
        "user_count": user_info["user_count"],
    }


def _user_team_json_generator(request, team, user):
    """Generate the json of team data"""
    team_id = team.id
    badges = request.db.get_badges_from_team(team_id)
    badges_count = len(badges)

    series = request.db.get_series_from_team(team_id)
    assertions = user.assertions
    assertion_ids = set([assertion.badge_id for assertion in assertions])

    series_info = []
    for elem in series:
        milestones_info = []
        milestones = elem.milestone
        for milestone in milestones:
            milestones_info.append(
                {
                    "milestone": milestone.as_dict(),
                    "series": elem.as_dict(),
                    "is_awarded": milestone.badge_id in assertion_ids,
                }
            )
        series_info.append(milestones_info)

    return {"badges_count": badges_count, "series_info": series_info}


@view_config(route_name="diff", renderer="diff.mak")
def diff(request):
    """Render user diff page."""

    # Get awarded assertions.
    if request.authenticated_userid:
        awarded_assertions = request.db.get_assertions_by_email(request.authenticated_userid)
    else:
        awarded_assertions = None

    user_a_id = request.matchdict.get("id_a")
    user_b_id = request.matchdict.get("id_b")
    user_a = _get_user(request, user_a_id)
    user_b = _get_user(request, user_b_id)

    if not user_a:
        raise HTTPNotFound(f"No such user {user_a_id!r}")
    if not user_b:
        raise HTTPNotFound(f"No such user {user_b_id!r}")

    if user_a.opt_out and user_a.email != request.authenticated_userid:
        raise HTTPNotFound(f"User {user_a_id!r} has opted out.")
    if user_b.opt_out and user_b.email != request.authenticated_userid:
        raise HTTPNotFound(f"User {user_b_id!r} has opted out.")

    # Get user badges.
    user_a_badges = [a.badge for a in user_a.assertions]
    user_b_badges = [a.badge for a in user_b.assertions]

    # Sort user badges by id.
    user_a_badges = sorted(user_a_badges, key=lambda badge: badge.id)
    user_b_badges = sorted(user_b_badges, key=lambda badge: badge.id)

    # Get total number of unique badges in the system.
    count_total_badges = len(request.db.get_all_badges().all())

    # Get percentage of badges earned.
    try:
        user_a_percent_earned = (float(len(user_a_badges)) / float(count_total_badges)) * 100
    except ZeroDivisionError:
        user_a_percent_earned = 0
    try:
        user_b_percent_earned = (float(len(user_b_badges)) / float(count_total_badges)) * 100
    except ZeroDivisionError:
        user_b_percent_earned = 0

    # Get rank. (same code found in leaderboard view function)
    user_a_rank = user_a.rank
    user_b_rank = user_b.rank
    user_count = request.db.session.query(m.Person).filter(m.Person.opt_out.is_(False)).count()

    try:
        user_a_percentile = (float(user_a_rank) / float(user_count)) * 100
    except ZeroDivisionError:
        user_a_percentile = 0
    try:
        user_b_percentile = (float(user_b_rank) / float(user_count)) * 100
    except ZeroDivisionError:
        user_b_percentile = 0

    # Diff badges.
    user_a_unique_badges = []
    user_b_unique_badges = []
    combined_badges = list(sorted(set(user_a_badges + user_b_badges), key=lambda badge: badge.id))
    shared_badges = []
    for badge in combined_badges:
        if badge in user_a_badges and badge not in user_b_badges:
            user_a_unique_badges.append(badge)
        elif badge in user_b_badges and badge not in user_a_badges:
            user_b_unique_badges.append(badge)
        elif badge in user_a_badges and badge in user_b_badges:
            shared_badges.append(badge)

    return dict(
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
        user_count=user_count,
        user_a=user_a,
        user_b=user_b,
        user_a_badges=user_a_badges,
        user_b_badges=user_b_badges,
        user_a_unique_badges=user_a_unique_badges,
        user_b_unique_badges=user_b_unique_badges,
        shared_badges=shared_badges,
        user_a_percent_earned=user_a_percent_earned,
        user_b_percent_earned=user_b_percent_earned,
        user_a_rank=user_a_rank,
        user_b_rank=user_b_rank,
        user_a_percentile=user_a_percentile,
        user_b_percentile=user_b_percentile,
    )


@view_config(route_name="user_team_json", renderer="json")
def user_team_json(request):
    """Render user team info as JSON dump."""

    user = _get_user(request, request.matchdict.get("id"))
    team = _get_team(request, request.matchdict.get("team_id"))

    if not user:
        request.response.status = "404 Not Found"
        return {"error": "No such user exists."}

    if user.opt_out and user.email != request.authenticated_userid:
        request.response.status = "404 Not Found"
        return {"error": "User has opted out."}

    return _user_team_json_generator(request, team, user)


@view_config(route_name="user_json", renderer="json")
def user_json(request):
    """Render user info JSON dump."""

    # So, here they can use their 'id' or their 'nickname'.
    # We'll try nickname first since we want to encourage that (or whatever)
    # and fall back to id if that fails.  If both fail, raise a 404.
    user = _get_user(request, request.matchdict.get("id"))

    if not user:
        request.response.status = "404 Not Found"
        return {"error": "No such user exists."}

    if user.opt_out and user.email != request.authenticated_userid:
        request.response.status = "404 Not Found"
        return {"error": "User has opted out."}

    return _user_json_generator(request, user)


@view_config(route_name="builder", renderer="builder.mak")
def builder(request):
    if request.authenticated_userid:
        awarded_assertions = request.db.get_assertions_by_email(request.authenticated_userid)
    else:
        awarded_assertions = None

    # set came_from so we can get back home after openid auth.
    request.session["came_from"] = request.route_url("builder")

    # get default creator field
    default_creator = None
    user = request.db.get_person(person_email=request.authenticated_userid)
    if user:
        default_creator = user.nickname or user.email

    badge_yaml = None
    if request.POST:
        token = request.session.get_csrf_token()
        if token != request.POST["csrf_token"]:
            raise HTTPForbidden("CSRF token did not match")
        badge_yaml = generate_badge_yaml(request.POST)

    return dict(
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
        default_creator=default_creator,
        badge_yaml=badge_yaml,
    )


@view_config(route_name="tags", renderer="tags.mak")
def tags(request):
    """Render tag page."""

    # Get awarded assertions.
    if request.authenticated_userid:
        awarded_assertions = request.db.get_assertions_by_email(request.authenticated_userid)
    else:
        awarded_assertions = None

    # Get badges matching tag.
    tags = [t.strip() for t in request.matchdict.get("tags").split(",")]
    if request.matchdict.get("match") == "all":
        badges = request.db.get_badges_from_tags(tags, match_all=True)
    else:
        badges = request.db.get_badges_from_tags(tags, match_all=False)

    return dict(
        tags=tags,
        badges=badges,
        auth_principals=request.effective_principals,
        awarded_assertions=awarded_assertions,
    )


@view_config(route_name="report", renderer="report.mak")
def report(request):
    """Render report page."""

    frame = "this week"

    start = get_start_week()
    stop = start + timedelta(days=6)

    user_to_rank = request.db._make_leaderboard(
        start=start,
        stop=stop,
    )

    return dict(
        auth_principals=request.effective_principals,
        user_to_rank=user_to_rank,
        start_date=start,
        stop_date=stop,
        frame=frame,
    )


@view_config(route_name="report_year", renderer="report.mak")
def report_year(request):
    """The leaderboard for a specific year."""

    frame = "year"

    ## TODO: how to make sure this doesn't break?
    year = int(request.matchdict.get("year"))

    start = date(year, 1, 1)
    stop = date(year, 12, 31)

    user_to_rank = request.db._make_leaderboard(
        start=start,
        stop=stop,
    )

    return dict(
        auth_principals=request.effective_principals,
        user_to_rank=user_to_rank,
        start_date=start,
        stop_date=stop,
        frame=frame,
    )


@view_config(route_name="report_year_month", renderer="report.mak")
def report_year_month(request):
    """The leaderboard for a specific month of a specific year."""

    frame = "month"

    ## TODO: how to make sure this doesn't break?
    year = int(request.matchdict.get("year"))
    month = int(request.matchdict.get("month"))

    now = datetime.now(timezone.utc)
    start = date(year, month, 1)
    if now.month == 12:
        stop = date(year + 1, 1, 1) - timedelta(days=1)
    else:
        stop = date(year, month + 1, 1) - timedelta(days=1)

    user_to_rank = request.db._make_leaderboard(
        start=start,
        stop=stop,
    )

    return dict(
        auth_principals=request.effective_principals,
        user_to_rank=user_to_rank,
        start_date=start,
        stop_date=stop,
        frame=frame,
    )


@view_config(route_name="report_year_month_day", renderer="report.mak")
def report_year_month_day(request):
    """The leaderboard for a specific month of a specific year."""

    frame = "day"

    ## TODO: how to make sure this doesn't break?
    year = int(request.matchdict.get("year"))
    month = int(request.matchdict.get("month"))
    day = int(request.matchdict.get("day"))

    start = date(year, month, day)
    stop = date(year, month, day) + timedelta(days=1)

    user_to_rank = request.db._make_leaderboard(
        start=start,
        stop=stop,
    )

    return dict(
        auth_principals=request.effective_principals,
        user_to_rank=user_to_rank,
        start_date=start,
        stop_date=stop,
        frame=frame,
    )


@view_config(route_name="report_year_week", renderer="report.mak")
def report_year_week(request):
    """The leaderboard for a specific week of a specific year."""

    frame = "week"

    ## TODO: how to make sure this doesn't break?
    year = int(request.matchdict.get("year"))
    week = int(request.matchdict.get("weeknumber"))

    # Get the week using the number of week
    start = date(year, 1, 1) + timedelta(weeks=week - 1)

    # Get the start of the week (as January 1st might not have been a Monday)
    start = get_start_week(start.year, start.month, start.day)
    stop = start + timedelta(days=6)

    user_to_rank = request.db._make_leaderboard(
        start=start,
        stop=stop,
    )

    return dict(
        auth_principals=request.effective_principals,
        user_to_rank=user_to_rank,
        start_date=start,
        stop_date=stop,
        frame=frame,
    )


@view_config(route_name="report_this_month")
def report_this_month(request):
    now = datetime.now(timezone.utc)
    year, month = now.year, now.month
    location = request.route_url("report_year_month", year=year, month=month)
    return HTTPFound(location=location)


@view_config(route_name="award_from_csv", permission="admin")
def award_from_csv(request):
    token = request.session.get_csrf_token()
    if token != request.POST["csrf_token"]:
        raise HTTPForbidden("CSRF token did not match")

    csv_file = request.POST["csv-file"].file
    successful_awards = 0
    """TODO: We need some validation here, and flash
    a message whether the awards were successful or not.
    This should be added at the same time that flash
    messages are added to the admin panel."""
    awards = dict()  # str(badge_id) : str(person_email)
    for line in csv_file:
        values = line.split(",")
        """Through the following if statement, it doesn't matter
        if the line has been entered as "email, badge" or
        "badge, email". The awards dict will be normalized
        to be "badge, email". This code assumes that one of the
        two values will be a valid email address."""
        if values[0].find("@") == -1:
            # If there is no @ sign in the first value, it is the badge id.
            awards[values[1].strip()] = values[0].strip()
        else:
            # If there is an @ sign in the first value, it is the email.
            awards[values[0].strip()] = values[1].strip()

    for email, badge_id in awards.items():
        # First, if the person doesn't exist, we automatically
        # create the person in this special case.
        if not request.db.person_exists(email=email):
            request.db.add_person(email)
        # Second, if the badge exists and the person has yet
        # to be awarded it, give it to them.
        if request.db.badge_exists(badge_id):
            if not request.db.assertion_exists(badge_id, email):
                # The None will default to datetime.now().
                request.db.add_assertion(badge_id, email, None)
                successful_awards += 1

    request.session.flash(f"Successfully awarded {successful_awards} badges.")
    return HTTPFound(location=request.route_url("admin"))


@view_config(context=str)
def html(context, request):
    return Response(context)


@view_config(context=m.Assertion, renderer="json")
def json(context, request):
    return context.as_dict()


@view_config(context="pyramid.httpexceptions.HTTPNotFound", renderer="404.mak")
def _404(request):
    request.response.status = 404
    return dict()


@view_config(context="pyramid.httpexceptions.HTTPServerError", renderer="500.mak")
def _500(request):
    request.response.status = 500
    return dict()


@view_config(route_name="assertion_widget", renderer="assertion_widget.mak")
def assertion_widget(request):
    person_id = request.matchdict.get("person")
    badge_id = request.matchdict.get("badge")
    user = request.db.get_person(id=person_id)
    if not user:
        raise HTTPNotFound(f"No such person {person_id!r}")

    def get_assertion():
        for assertion in user.assertions:
            if assertion.badge.id == badge_id:
                return assertion
        raise HTTPNotFound("User does not have that badge")

    assertion = get_assertion()
    return dict(assertion=assertion)


def modify_rst(rst):
    """Downgrade some of our rst directives if docutils is too old."""

    try:
        # The rst features we need were introduced in this version
        minimum = [0, 9]
        version = [int(elem) for elem in docutils.__version__.split(".")]

        # If we're at or later than that version, no need to downgrade
        if version >= minimum:
            return rst
    except Exception:
        # If there was some error parsing or comparing versions, run the
        # substitutions just to be safe.
        pass

    # Otherwise, make code-blocks into just literal blocks.
    substitutions = {
        ".. code-block:: javascript": "::",
    }
    for old, new in substitutions.items():
        rst = rst.replace(old, new)

    return rst


def modify_html(html):
    """Perform style substitutions where docutils doesn't do what we want."""

    substitutions = {
        '<tt class="docutils literal">': "<code>",
        "</tt>": "</code>",
    }
    for old, new in substitutions.items():
        html = html.replace(old, new)

    return html


def _load_docs(directory, endpoint):
    """Utility to load an RST file and turn it into fancy HTML."""

    fname = os.path.join(directory, endpoint + ".rst")
    with codecs.open(fname, "r", "utf-8") as f:
        rst = f.read()

    rst = modify_rst(rst)

    api_docs = docutils.examples.html_body(rst)

    api_docs = modify_html(api_docs)

    api_docs = markupsafe.Markup(api_docs)
    return api_docs


htmldocs = {}


def load_docs(request, key):
    possible_keys = ["about", "footer"]

    # Load from disk only once on first request.
    if not htmldocs:
        here = os.path.dirname(os.path.abspath(__file__))
        dflt = os.path.join(here, "sitedocs")
        directory = request.registry.settings.get("tahrir.sitedocs_dir", dflt)
        for k in possible_keys:
            htmldocs[k] = _load_docs(directory, k)

    if key not in htmldocs:
        raise KeyError(f"{key!r} is not permitted.")

    return htmldocs[key]


def get_start_week(year=None, month=None, day=None):
    """For a given date, retrieve the day the week started.
    For any missing parameters (ie: None), use the value of the current
    day.

    :kwarg year: year to consider when searching a week.
    :kwarg month: month to consider when searching a week.
    :kwarg day: day to consider when searching a week.
    :return a Date of the day the week started either based on the
        current utc date or based on the information.
    """
    now = datetime.now(timezone.utc)
    if not year:
        year = now.year
    if not month:
        month = now.month
    if not day:
        day = now.day
    week_day = date(year, month, day)
    week_start = week_day - timedelta(days=week_day.weekday())
    return week_start



import os
from configparser import ConfigParser

from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.config import Configurator
from pyramid.session import SignedCookieSessionFactory
from pyramid.settings import asbool
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from tahrir_api.dbapi import TahrirDatabase

from . import notifications
from .app import get_root
from .utils import cache


def main(global_config, **settings):
    """This function returns a Pyramid WSGI application."""

    session_cls = scoped_session(
        sessionmaker(
            bind=create_engine(settings["sqlalchemy.url"]),
        )
    )

    def get_db(request):
        """Database retrieval function to be added to the request for
        calling anywhere.
        """
        session = session_cls()
        return TahrirDatabase(
            session=session, autocommit=True, notification_callback=notifications.callback
        )

    required_keys = [
        "tahrir.pngs.uri",
        "tahrir.admin",
        "tahrir.title",
        "tahrir.base_url",
    ]

    # validate the config
    for key in required_keys:
        if key not in settings:
            raise ValueError(f"{key} required in settings.")

    # Make data dir if it doesn't already exist.
    settings["tahrir.pngs.uri"] = os.path.abspath(settings["tahrir.pngs.uri"])
    if not os.path.exists(settings["tahrir.pngs.uri"]):
        os.makedirs(settings["tahrir.pngs.uri"])

    # Load secret stuff from secret.ini.
    try:
        default_path = os.path.abspath("secret.ini")
        secret_path = settings.get("tahrir.secret_config_path", default_path)
        # TODO: There is a better way to log this message than print.
        print(f"Reading secrets from {secret_path!r}")
        parser = ConfigParser()
        parser.read(secret_path)
        secret_config = dict(parser.items("tahrir"))
        settings.update(secret_config)
    except Exception as e:
        # TODO: There is a better way to log this message than print.
        print(f"Failed to load secret.ini.  Reason: {str(e)!r}")

    authn_policy = AuthTktAuthenticationPolicy(
        secret=settings["authnsecret"],
        callback=groupfinder,  # groupfinder callback checks for admin privs
        hashalg="sha512",  # because md5 is deprecated
        secure=asbool(settings.get("tahrir.secure_cookies", True)),
        http_only=asbool(settings.get("tahrir.httponly_cookies", True)),
    )
    authz_policy = ACLAuthorizationPolicy()
    session_factory = SignedCookieSessionFactory(
        secret=settings["session.secret"],
        secure=asbool(settings.get("tahrir.secure_cookies", True)),
        httponly=asbool(settings.get("tahrir.httponly_cookies", True)),
    )

    # Configure our cache that we instantiated earlier.
    cache.configure_from_config(settings, "dogpile.cache.")

    config = Configurator(
        settings=settings,
        root_factory=get_root,
        session_factory=session_factory,
        authentication_policy=authn_policy,
        authorization_policy=authz_policy,
    )

    config.include("pyramid_mako")

    config.add_request_method(get_db, "db", reify=True)

    config.add_static_view(
        "static",
        settings.get("tahrir.theme_name", "tahrir") + ":static",
        cache_max_age=3600,
    )
    config.add_static_view(
        "pngs",
        settings["tahrir.pngs.uri"],
        cache_max_age=3600,
    )

    config.add_route("home", "/")
    config.add_route("heartbeat", "/heartbeat")

    # main admin endpoint
    config.add_route("admin", "/admin")

    # delegated admin endpoints
    config.add_route("award", "/award")
    config.add_route("invite", "/invite")
    config.add_route("add_tag", "/add_tag")

    config.add_route("qrcode", "/qrcode")
    config.add_route("badge", "/badge/{id}")
    config.add_route("badge_full", "/badge/{id}/full")
    config.add_route("badge_json", "/badge/{id}/json")
    config.add_route("badge_rss", "/badge/{id}/rss")
    config.add_route("badge_stl", "/badge/{id}/stl")
    config.add_route("builder", "/builder")
    config.add_route("about", "/about")
    config.add_route("explore", "/explore")
    config.add_route("explore_badges", "/explore/badges")
    config.add_route("explore_badges_rss", "/explore/badges/rss")
    config.add_route("leaderboard", "/leaderboard")
    config.add_route("leaderboard_json", "/leaderboard/json")
    config.add_route("rank_json", "/leaderboard/{id}/json")
    config.add_route("tags", "/tags/{tags}/{match}")
    config.add_route("user", "/user/{id}")
    config.add_route("user_json", "/user/{id}/json")
    config.add_route("user_rss", "/user/{id}/rss")
    config.add_route("user_team_json", "/user/{id}/team/{team_id}/json")
    config.add_route("diff", "/diff/{id_a}/{id_b}")
    config.add_route("report", "/report")
    config.add_route("report_this_month", "/report/this/month")
    config.add_route("report_year", "/report/{year}")
    config.add_route("report_year_month", "/report/{year}/{month}")
    config.add_route("report_year_week", "/report/{year}/week/{weeknumber}")
    config.add_route("report_year_month_day", "/report/{year}/{month}/{day}")
    config.add_route("award_from_csv", "/award_from_csv")

    config.include("tahrir.auth")

    # config.add_route('login', '/login')
    # config.add_route('logout', '/logout')

    # Used to grab a "was awarded" html snippet asynchronously
    config.add_route("assertion_widget", "/_w/assertion/{person}/{badge}")

    config.scan()

    return config.make_wsgi_app()


# I think this is fine here...
def groupfinder(userid, request):
    """Currently, this function simply checks if the user
    is listed as an admin in the config file (tahrir.ini).
    This is the callback function used by the authorization
    policy."""
    admins = [admin.strip() for admin in request.registry.settings["tahrir.admin"].split(",")]
    if userid in admins:
        return ["group:admins"]
    else:
        return []



import logging
import shutil
import tempfile

log = logging.getLogger(__name__)


def scale_to_standard_size(filename):
    try:
        import magickwand.image as magick
    except Exception as e:
        log.warn(str(e))
        log.warn("Did not scale image to standard size")
        return

    img = magick.Image(filename)
    w, h = img.size
    s = min(w, h)
    wo = int(float(w - s) / 2.0)
    ho = int(float(h - s) / 2.0)

    img.crop((s, s), (wo, ho))
    img.thumbnail(256)
    handle, tempname = tempfile.mkstemp(suffix=".png")
    img.save(tempname)
    shutil.move(filename, filename + ".original")
    shutil.move(tempname, filename)



import logging
import sys
import traceback

import backoff
import pyramid.threadlocal
from fedora_messaging import api as fm_api
from fedora_messaging import exceptions as fm_exceptions
from pyramid.settings import asbool

log = logging.getLogger(__name__)


def _backoff_hdlr(details):
    log.warning(f"Publishing message failed. Retrying. {traceback.format_tb(sys.exc_info()[2])}")


@backoff.on_exception(
    backoff.expo,
    (fm_exceptions.ConnectionException, fm_exceptions.PublishException),
    max_tries=3,
    on_backoff=_backoff_hdlr,
)
def _publish(message):
    fm_api.publish(message)


def callback(message):
    request = pyramid.threadlocal.get_current_request()
    settings = request.registry.settings
    if asbool(settings.get("tahrir.use_fedmsg", False)):
        log.debug(f"Publishing fedoramessage `{message}`")
        try:
            _publish(message)
        except fm_exceptions.BaseException:
            log.error(
                f"Publishing message failed. Giving up. {traceback.format_tb(sys.exc_info()[2])}"
            )



"""This module contains generic Pyramid integration with Authlib >= 1.0.

It could be sent to Authlib upstream.
"""

from authlib.integrations.base_client import (
    BaseApp,
    BaseOAuth,
    FrameworkIntegration,
    OAuth1Mixin,
    OAuth2Mixin,
    OAuthError,
    OpenIDMixin,
)
from authlib.integrations.requests_client import OAuth1Session, OAuth2Session

from .oauth import PyramidAppMixin, PyramidIntegrationMixin


class PyramidIntegration(PyramidIntegrationMixin, FrameworkIntegration):
    """Pyramid-specific framework integration."""


class PyramidAppMixinAuthlib1(PyramidAppMixin):
    """Generic methods for a Pyramid Authlib app running on Authlib 1.x."""

    def save_authorize_data(self, request, **kwargs):
        """Save the authorization data in the session."""
        state = kwargs.pop("state", None)
        if state:
            self.framework.set_state_data(request.session, state, kwargs)
        else:
            raise RuntimeError("Missing state value")


class PyramidOAuth1App(PyramidAppMixinAuthlib1, OAuth1Mixin, BaseApp):
    """OAuth1 Authlib application for Pyramid."""

    client_cls = OAuth1Session

    def authorize_access_token(self, request, **kwargs):
        """Fetch access token in one step.

        :param request: HTTP request instance from Django view.
        :return: A token dict.
        """
        params = dict(request.GET)
        state = params.get("oauth_token")

        if not state:
            raise OAuthError(description='Missing "oauth_token" parameter')

        data = self.framework.get_state_data(request.session, state)
        if not data:
            raise OAuthError(description='Missing "request_token" in temporary data')

        params["request_token"] = data["request_token"]
        params.update(kwargs)
        self.framework.clear_state_data(request.session, state)
        return self.fetch_access_token(**params)


class PyramidOAuth2App(PyramidAppMixinAuthlib1, OAuth2Mixin, OpenIDMixin, BaseApp):
    """OAuth2 Authlib application for Pyramid."""

    client_cls = OAuth2Session

    def authorize_access_token(self, request, **kwargs):
        """Fetch access token in one step.

        :param request: HTTP request instance from Pyramid view.
        :return: A token dict.
        """
        if request.method == "GET":
            error = request.GET.get("error")
            if error:
                self.framework.clear_state_data(request.session, request.GET.get("state"))
                description = request.GET.get("error_description")
                raise OAuthError(error=error, description=description)
            params = {
                "code": request.GET.get("code"),
                "state": request.GET.get("state"),
            }
        else:
            params = {
                "code": request.POST.get("code"),
                "state": request.POST.get("state"),
            }

        state_data = self.framework.get_state_data(request.session, params.get("state"))
        self.framework.clear_state_data(request.session, params.get("state"))
        params = self._format_state_params(state_data, params)
        token = self.fetch_access_token(**params, **kwargs)

        if "id_token" in token and "nonce" in state_data:
            userinfo = self.parse_id_token(token, nonce=state_data["nonce"])
            token["userinfo"] = userinfo
        return token


class OAuth(BaseOAuth):
    """A Pyramid-specific OAuth app."""

    oauth1_client_cls = PyramidOAuth1App
    oauth2_client_cls = PyramidOAuth2App
    framework_integration_cls = PyramidIntegration



"""Views related to authentication."""

import typing

from authlib.integrations.base_client import OAuthError
from pyramid.httpexceptions import HTTPFound, HTTPUnauthorized
from pyramid.security import forget

from .utils import get_and_store_user, get_final_redirect

if typing.TYPE_CHECKING:  # pragma: no cover
    import pyramid.request.Request  # noqa: 401
    import pyramid.response.Response  # noqa: 401


def logout(request: "pyramid.request.Request") -> HTTPFound:
    """
    Log out the user.

    Args:
        request: The current request, which is used to remove the user's
            authentication cookies.
    Returns:
        A 302 redirect to the home page.
    """
    headers = forget(request)
    return HTTPFound(location=request.route_url("home"), headers=headers)


def login_oidc(request: "pyramid.request.Request"):
    """Send the user to the OpenID Connect provider to log in.

    Args:
        request (pyramid.request.Request): The Pyramid request.

    Returns:
        pyramid.response.Response: A redirect to the OIDC provider's login frame.
    """
    provider = request.registry.oidc.create_client("fedora")
    redirect_uri = request.route_url("oidc_authorize")
    return provider.authorize_redirect(request, redirect_uri)


def authorize_oidc(request: "pyramid.request.Request"):
    """Verify the response from the OpenID Connect provider and log the user in.

    Args:
        request (pyramid.request.Request): The Pyramid request.

    Returns:
        pyramid.response.Response: A redirection to the previously visited page.
    """
    try:
        token = request.registry.oidc.fedora.authorize_access_token(request)
    except OAuthError as e:
        raise HTTPUnauthorized(f"Authentication failed: {e.description}") from e
    response = get_final_redirect(request)
    get_and_store_user(request, token["access_token"], response)
    return response



"""Constants that do not need to be configurable."""

SCOPES = " ".join(
    [
        "openid",
        "email",
        "profile",
        "https://id.fedoraproject.org/scope/groups",
        "https://id.fedoraproject.org/scope/agreements",
    ]
)



"""This module sets up OpenID Connect authentication"""

from .constants import SCOPES
from .fedora import FedoraApp
from .oauth_1 import OAuth


def includeme(config):
    """Set up the authentication."""
    oauth = OAuth()
    oauth.register(
        "fedora",
        client_kwargs={
            "scope": SCOPES,
            "token_endpoint_auth_method": "client_secret_post",
        },
        client_cls=FedoraApp,
    )
    config.registry.oidc = oauth

    config.add_route("logout", "/logout")
    config.add_view("tahrir.auth.views.logout", route_name="logout")
    config.add_route("login", "/login")
    config.add_view("tahrir.auth.views.login_oidc", route_name="login")
    config.add_route("oidc_authorize", "/oidc/authorize")
    config.add_view("tahrir.auth.views.authorize_oidc", route_name="oidc_authorize")
